/*
 * <3BSD notice header>
 *
 * Copyright (C) 2003-2008 Raritan, Inc. <http://www.raritan.com/>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 * 
 *     * Neither the name of [original copyright holder] nor the names of
 *       its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written
 *       permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * w83793g_sensor.h
 *
 * "Super class" for all Winbond W83793G sensors
 * 
 * (c) 2006 Supermicro, 2006/04/04, Edward Pan<edwardp@supermicro.com>
 * (c) 2006 ASUSTek, 2006/10/05, JinCheng
 */

#include <malloc.h>

#include <pp/base.h>
#include <pp/bmc/bmc_config.h>
#include <pp/bmc/ipmi_err.h>
#include <pp/bmc/ipmi_sdr.h>
#include <pp/bmc/ipmi_bits.h>
#include <pp/bmc/ipmi_cmd_sensor.h>

#include <pp/bmc/sensor.h>
#include <pp/bmc/ipmi_sdr.h>
#include <pp/bmc/topo_base_obj.h>
#include <pp/bmc/topo_factory.h>
#include <pp/bmc/tp_i2c_chip.h>
#include <pp/bmc/tp_scan_sensdev.h>

#include "drivers/w83793g.h"
#include "drivers/w83793g_sensor.h"


/*
 * private structures
 ************************/

/*
 * W83793G sensor's structure
 */

/*
typedef enum{
    VOLTAGE_CoreA = 1,
    VOLTAGE_CoreB,
    VOLTAGE_12VSEN,
    VOLTAGE_VSEN2_1V5,
    VOLTAGE_3VSEN,
    VOLTAGE_VBAT,
    VOLTAGE_5VDD,
    VOLTAGE_5VSB,
    VOLTAGE_VSEN1_N12V,
} voltage_pin_t;
*/

typedef struct sensor_config_s {
    /* the sensor type */
    unsigned char sensor_type;
    unsigned char reading_type;
    unsigned char sdr_type;
    unsigned char unit_type;
    unsigned char analog_type;
    /* update function */
    void (*update)(pp_sensor_scannable_t*);
    /* configuration */
    union {
        struct {
            unsigned char reg;
//            unsigned char volt_pin;
        } volt;
        struct {
            unsigned char reg;
        } fan;
        struct {
            unsigned char reg;
        } temp;
        struct {
            unsigned char reg;
            unsigned char bit;
        } intr;
    } config; 
    /* sdr configuration */
    unsigned short sdr_mtol;
    unsigned short sdr_bacc;
    char sdr_bexp;
    char sdr_rexp;
    unsigned short retry_cnt;
} sensor_config_t;

typedef struct {
    /* the basic sensor description, must be first */
    pp_tp_scan_sensdev_t base;
    
    pp_tp_i2c_chip_t* chip;

    /* sensor configuration, pointer to statically allocated memory */
    sensor_config_t* cfg;

} w83793g_sensor_t;

/*
 * private function prototypes, used by all W83793G sensors
 *************************************************************/

static w83793g_sensor_t* w83793g_sensor_create(/* globals */
					      const char* id,
                                              pp_tp_i2c_chip_t* chip,
                                              /* the sensor idx */
                                              int sensor_idx,
					      pp_tp_cond_t* scan);
//static char* w83793g_to_string(pp_tp_sensdev_t* sensor);
static ipmi_sdr_header_t* w83793g_default_sdr(pp_tp_sensdev_t* s);
static void w83793g_update_voltage_reading(pp_sensor_scannable_t* s);
static void w83793g_update_tacho_reading(pp_sensor_scannable_t* s);
static void w83793g_update_temp_reading(pp_sensor_scannable_t* s);
static void w83793g_update_intr_reading(pp_sensor_scannable_t* s);
static void w83793g_destroy(w83793g_sensor_t* sensor);

/**
 * The actual configurations of the W83793G sensors
 *****************************************************************/
static sensor_config_t sensor_config[] = {
    { /* 0: System Temperature 1(Front)		[THR1] */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = w83793g_update_temp_reading,
        .config = { .temp = { .reg = 0x20, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
    { /* 1: System Temperature 2(Rear)		[THR2] */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = w83793g_update_temp_reading,
        .config = { .temp = { .reg = 0x21, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
    { /* 2: CPU1_core1(CPU1) Temperature	[1_D+/-] */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = w83793g_update_temp_reading,
        .config = { .temp = { .reg = 0x1C, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
    { /* 3: CPU1_core2(CPU3) Temperature	[3_D+/-] */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = w83793g_update_temp_reading,
        .config = { .temp = { .reg = 0x1E, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
    { /* 4: CPU2_core1(CPU2) Temperature	[2_D+/-] */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = w83793g_update_temp_reading,
        .config = { .temp = { .reg = 0x1D, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
    { /* 5: CPU2_core2(CPU4) Temperature	[4_D+/-] */
        .sensor_type = IPMI_SENSOR_TYPE_TEMPERATURE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_DEGREES_C,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_2_COMPL,
	.update = w83793g_update_temp_reading,
        .config = { .temp = { .reg = 0x1F, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(1, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
    { /* 6: Vcore 1							[VcoreA] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_CORE1, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(8, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 7: Vcore 2							[VcoreB] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_CORE2, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(8, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 8: Vtt								[VTT] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_VTT, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(8, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 9: +1.5V/Vcore 3/Mem_P1			[VSEN1] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_1_5V, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 10: +1.8V/Vcore 4/Mem_P2			[VSEN2] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_1_8V, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 11: +3.3V							[+3VSEN] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_3V, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 12: +12V							[+12VSEN] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_12V, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(96, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 13: +5V							[5VDD] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_5V, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(24, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 14: +5VSB							[5VSB] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_5VSB, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(24, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 15: VBAT							[VBAT] */
        .sensor_type = IPMI_SENSOR_TYPE_VOLTAGE,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_VOLTS,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
	.update = w83793g_update_voltage_reading,
        .config = { .volt = { .reg = W83793G_VOLTAGE_VBAT, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(16, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = -3,
        .retry_cnt = 0,
	},
	{ /* 16: FAN1(CPU_FAN1)					[FANIN1] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_RPM,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x23, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
	{ /* 17: FAN2(CPU_FAN2)					[FANIN2] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_RPM,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x25, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
	{ /* 18: FAN3(FRONT_FAN1)				[FANIN3] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_RPM,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x27, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
	{ /* 19: FAN4(FRONT_FAN2)				[FANIN4] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_RPM,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x29, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
	{ /* 20: FAN5(FRONT_FAN3)				[FANIN5] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_RPM,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x2B, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
	{ /* 21: FAN6(FRONT_FAN4)				[FANIN6] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_RPM,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x2D, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
	{ /* 22: FAN7(REAR_FAN1)				[FANIN7] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
	.reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
	.sdr_type = IPMI_SDR_FULL_SENSOR,
	.unit_type = IPMI_UNIT_TYPE_RPM,
	.analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x2F, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
	.sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
	.sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
	},
	{ /* 23: FAN8(REAR_FAN2)				[FANIN8] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
        .reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
        .sdr_type = IPMI_SDR_FULL_SENSOR,
        .unit_type = IPMI_UNIT_TYPE_RPM,
        .analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x31, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
        .sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
        .sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
    },
	{ /* 24: FAN9(FBD_FAN1)(FRONT_FAN5)		[FANIN9] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
        .reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
        .sdr_type = IPMI_SDR_FULL_SENSOR,
        .unit_type = IPMI_UNIT_TYPE_RPM,
        .analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x33, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
        .sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
        .sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
    },
	{ /* 25: FAN10(FBD_FAN2)(FRONT_FAN6)	[FANIN10] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
        .reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
        .sdr_type = IPMI_SDR_FULL_SENSOR,
        .unit_type = IPMI_UNIT_TYPE_RPM,
        .analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x35, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
        .sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
        .sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
    },
	{ /* 26: FAN11(CPU_FAN3)				[FANIN11] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
        .reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
        .sdr_type = IPMI_SDR_FULL_SENSOR,
        .unit_type = IPMI_UNIT_TYPE_RPM,
        .analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x37, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
        .sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
        .sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
    },
	{ /* 27: FAN12(CPU_FAN4)				[FANIN12] */
        .sensor_type = IPMI_SENSOR_TYPE_FAN,
        .reading_type = IPMI_EVENT_READING_TYPE_THRESHOLD,
        .sdr_type = IPMI_SDR_FULL_SENSOR,
        .unit_type = IPMI_UNIT_TYPE_RPM,
        .analog_type = IPMI_ANALOG_DATA_FORMAT_UNSIGNED,
        .update = w83793g_update_tacho_reading,
        .config = { .fan = { .reg = 0x39, }, },
        .sdr_mtol = PP_BMC_SDR_M_TOL_2_MTOL(80, 0),
        .sdr_bacc = PP_BMC_SDR_B_ACCURACY_2_BACC(0, 0),
        .sdr_bexp = 0, .sdr_rexp = 0,
        .retry_cnt = 0,
    },
    { /* 28: Chassis intrusion				[CASEOPEN#] */
//        .sensor_type = IPMI_SENSOR_TYPE_PHYSICAL_SECURITY,
        .sensor_type = 0xc2,
	.reading_type = IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC,
	.sdr_type = IPMI_SDR_COMPACT_SENSOR,
	.update = w83793g_update_intr_reading,
        .config = { .intr = { .reg = 0x4E, .bit = 6, }, },
        .retry_cnt = 0,
    }
};

/*
 * implementations
 * -------------------------------
 */

pp_tp_obj_t* pp_sensor_w83793g_sens_ctor(const char* id, vector_t* args) {
    const char* fn = __FUNCTION__;
    const char* errmsg = "%s: '%s' failed: %s";
    pp_strstream_t err = PP_STRSTREAM_INITIALIZER;
    w83793g_sensor_t* o = NULL;
    pp_tp_i2c_chip_t* chip;
    pp_tp_cond_t* scan_cond;
    int snum;

    if (pp_tp_arg_scanf(args, 0, &err, "o<h>d|o<sc>",
			&chip, &snum, &scan_cond) != 3) {
	pp_bmc_log_perror(errmsg, fn, id, pp_strstream_buf(&err));
    } else if (strcmp(chip->model, W83793G_MODEL_STRING)) {
	errno = EINVAL;
	pp_bmc_log_perror(errmsg, fn, id, "incorrect chip model");
    } else if (snum > PP_SENSOR_W83793G_MAX_SENSOR_NUM) {
	pp_bmc_log_perror(errmsg, fn, id, "invalid sensor number");
    } else {
        o = w83793g_sensor_create(id, chip, snum, scan_cond);
    }
    pp_strstream_free(&err);
    return (pp_tp_obj_t*)o;
}

/*
 * private functions
 ********************************************************************/

static w83793g_sensor_t* w83793g_sensor_create(/* globals */
					      const char* id,
					      pp_tp_i2c_chip_t* chip,
                                              /* the sensor idx */
                                              int sensor_idx,
					      pp_tp_cond_t* scan)
{
    sensor_config_t *cfg = &sensor_config[sensor_idx];
    w83793g_sensor_t* wbs = malloc(sizeof(w83793g_sensor_t));

    /* globals */
    pp_tp_scan_sensdev_init(&wbs->base, PP_TP_SCAN_SENS_DEV, id,
			    (void(*)(pp_tp_obj_t*))w83793g_destroy,
			    w83793g_default_sdr, cfg->update, scan);
    wbs->chip = pp_tp_i2c_chip_duplicate(chip);
    wbs->cfg = cfg;

    /* register for scanning  */
    pp_tp_scan_sensdev_register(&wbs->base, 0);

    return (w83793g_sensor_t*)wbs;
}

static ipmi_sdr_header_t* w83793g_default_sdr(pp_tp_sensdev_t* s) {
    w83793g_sensor_t* this = (w83793g_sensor_t*)s;
    sensor_config_t* cfg = this->cfg;
    ipmi_sdr_header_t* sdr = NULL;

    if (cfg->sdr_type == IPMI_SDR_FULL_SENSOR) {
	sdr = (ipmi_sdr_header_t*)ipmi_sdr_full_part_create(
	    cfg->sensor_type, cfg->reading_type, cfg->unit_type, cfg->analog_type,
	    cfg->sdr_mtol, cfg->sdr_bacc, cfg->sdr_bexp, cfg->sdr_rexp,
	    ipmi_sdr_max_by_form(cfg->analog_type),
	    ipmi_sdr_min_by_form(cfg->analog_type));
    } else if (cfg->sdr_type == IPMI_SDR_COMPACT_SENSOR) {
	// could be a compact SDR entrie, but the Supermicro client only supports full SDR records!
//	sdr = (ipmi_sdr_header_t*)ipmi_sdr_compact_create(cfg->sensor_type, cfg->reading_type);
	sdr = (ipmi_sdr_header_t*)ipmi_sdr_full_part_create(
	    cfg->sensor_type, cfg->reading_type, 0, 0, 0, 0, 0, 0, 0, 0);
    } else {
	pp_bmc_log_error("w83793g_default_sdr failed: SDR type %d not supported", cfg->sdr_type);
    }

    return sdr;
}

static void w83793g_update_voltage_reading(pp_sensor_scannable_t* s) {
    w83793g_sensor_t* wbs = PP_TP_SCANNABLE_2_OBJ_CAST(s, w83793g_sensor_t);
    sensor_config_t* cfg = wbs->cfg;
    int ret = -2;
    pp_tp_i2c_chip_t* chip = wbs->chip;
    u_char value_mv, value;
    int fail_ret = 0;
//    int really_mv;

    if (pp_tp_i2c_chip_is_initialized(chip)
	&& pp_tp_scan_sensdev_do_scan(&wbs->base)
	&& (PP_SUC == (ret = pp_tp_i2c_chip_pre_com(chip)))) {

	ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.volt.reg,
					&value_mv);
	pp_tp_i2c_chip_post_com(chip);
	
	if (ret == PP_SUC) {
            /*
	    switch(cfg->config.volt.volt_pin){
            case VOLTAGE_CoreA:				// offset: 0x10
	         pp_bmc_log_info("CoreA V = %d", value_mv);    		
            case VOLTAGE_CoreB:                         // offset: 0x11
                 pp_bmc_log_info("CoreB V = %d", value_mv);    
                 really_mv = (float)value_mv* 0.008;
                 break;
            case VOLTAGE_12VSEN:
		 pp_bmc_log_info("12V = %d", value_mv);    
                 really_mv = (float)value_mv * 0.096;   // offset: 0x17
                 break;
            case VOLTAGE_VSEN2_1V5:                     // offset: 0x15
                 pp_bmc_log_info("1.5V = %d", value_mv);     
            case VOLTAGE_3VSEN:                         // offset: 0x16
                 pp_bmc_log_info("3.3V = %d", value_mv);    
            case VOLTAGE_VBAT:                          // offset: 0x1A
                 pp_bmc_log_info("BAT V= %d", value_mv);      
                 really_mv = (float)value_mv * 0.016;
                 break;
            case VOLTAGE_5VDD:                          // offset: 0x18
                  pp_bmc_log_info("5VDD = %d", value_mv);    
            case VOLTAGE_5VSB:                          // offset: 0x19
                  pp_bmc_log_info("5VSB = %d", value_mv);    
                 really_mv = (float)value_mv * 0.024;  
                 break;
            case VOLTAGE_VSEN1_N12V:
                 pp_bmc_log_info("-12V = %d", value_mv);    
                 really_mv = ((float)value_mv - 114.22)/6.73;
                 pp_bmc_log_info("-12V = %f", really_mv);    
                 value_mv = (u_char)really_mv; 
                 break;
            default:
                  ret = value_mv;
             
            
            }*/
             
	      	
             ret = value_mv;
	     value=value_mv;

	      if ( value_mv == 0x00 || value_mv == 0xff || ret == 0 )	// fixed bug for reset
	      {
		    fail_ret = 1;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_CORE1 && ( value>=0xde || value <=0x58 ) )	// Vcore
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_CORE2 && ( value>=0xde || value <=0x58) )	// 3.3V
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_VTT && ( value>=0xb5 || value <=0x7d) )	// 5V
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_1_5V && ( value>=0x67 || value <=0x48) )	// 12V
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_1_8V && ( value>=0x81 || value <=0x5f) )	// 5VSB
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_3V && ( value>=0xe2 || value <=0xaf ) )	// VBAT
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_12V && ( value>=0x86 || value <=0x71 ) )	// VBAT
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_5V && ( value>=0xe5 || value <=0xb1 ) )	// VBAT
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_5VSB && ( value>=0xe5 || value <=0xb1 ) )	// VBAT
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( cfg->config.volt.reg == W83793G_VOLTAGE_VBAT && ( value>=0xce || value <=0x9f ) )	// VBAT
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else if ( value_mv <=0x40 )	// fixed bug for reset
	      {
		    cfg->retry_cnt++;
		    if ( cfg->retry_cnt == 0)
		       cfg->retry_cnt = 10;
	      }
	      else
	      {
		 cfg->retry_cnt=0;
	      }
	      
	      if ( fail_ret == 1 || (cfg->retry_cnt>0 && cfg->retry_cnt <= 3 ) )
		 ret = -2;

 
//             if ( ret == 0 )
//                ret = -2;
            //ret = (u_char)really_mv;
//             if(cfg->config.volt.volt_pin == VOLTAGE_VSEN1_N12V){   
//                 really_mv =(0 - (((int)value_mv * 1.48) - 169.2));
//                 ret = really_mv;
//	      }
     
	   }
  	   else
	      ret = -2;

        }
    pp_tp_scan_sensdev_update_reading(&wbs->base, ret);
}

static void w83793g_update_tacho_reading(pp_sensor_scannable_t* s) {
    w83793g_sensor_t* wbs = PP_TP_SCANNABLE_2_OBJ_CAST(s, w83793g_sensor_t);
    sensor_config_t* cfg = wbs->cfg;
    int ret = PP_ERR;
    pp_tp_i2c_chip_t* chip = wbs->chip;
    u_char low, high;
    u_int value;
//    u_int pulses_per_minute;

    if (pp_tp_i2c_chip_is_initialized(chip)
	&& pp_tp_scan_sensdev_do_scan(&wbs->base)
	&& (PP_SUC == pp_tp_i2c_chip_pre_com(chip))) {

        // read the configuration and fan registers
	ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.fan.reg, &high);
	ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.fan.reg + 1, &low);

        pp_tp_i2c_chip_post_com(chip);

        if (ret == PP_SUC) {
            value = (((high & 0x0f) << 8) | (low));
             
//            if (value == 0xffff || value == 0) {
            if (value >= 0x0ff0 || value <= 65) {
                // fan is not moving, set value to 0 and return
                ret = 0;
		cfg->retry_cnt++;
		if ( cfg->retry_cnt == 0)
		   cfg->retry_cnt = 10;
            } else {
       		cfg->retry_cnt =0;
                // the counter tells us the number of clock cycles (90 kHz)
		// for one fan revolution
//                pulses_per_minute = 1350000 / value;

                // convert the count into a byte value
//                ret = min((u_int)255, pulses_per_minute / 50); // why pulses_per_minutes / 50 ?
//                ret = min((u_int)255, pulses_per_minute / 80); // why pulses_per_minutes / 50 ?
					ret = (1350000 / value) / 80;
            }


	    if ( ret == 0 )
	    {
	       if ( cfg->retry_cnt>0 && cfg->retry_cnt<=3 )		// fixed bug for divisor fail
	          ret = -2; 
	    }
	    else if (ret <= 0x05 )
	    {
	       if ( cfg->retry_cnt>0 && cfg->retry_cnt<=5 )		// fixed bug for divisor fail
	          ret = -2; 
	    }

        }
	else
	   ret = -2;
    }
    pp_tp_scan_sensdev_update_reading(&wbs->base, ret);
}

static void w83793g_update_temp_reading(pp_sensor_scannable_t* s) {
    w83793g_sensor_t* wbs = PP_TP_SCANNABLE_2_OBJ_CAST(s, w83793g_sensor_t);
    sensor_config_t* cfg = wbs->cfg;
    int ret = PP_ERR;
    pp_tp_i2c_chip_t* chip = wbs->chip;
    u_char value;

    if (pp_tp_i2c_chip_is_initialized(chip)
	&& pp_tp_scan_sensdev_do_scan(&wbs->base)
	&& (PP_SUC == pp_tp_i2c_chip_pre_com(chip))) {

        ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.temp.reg,
					  &value);
        pp_tp_i2c_chip_post_com(chip);

        if (ret == PP_SUC) {
            // TODO: check for correct values --> 0x80 means no value
            if (value == 0x80) {
                errno = ENODATA;
//                ret = PP_ERR;
//            } else {
//                ret = value;
            }
            ret = value;

            if ( ret == 0 )
               ret = -2;

	    if ( ret >= 0x80 )	
            {	
	       ret = -2;
	    }
        }
	else
	   ret = -2;
    }
    pp_tp_scan_sensdev_update_reading(&wbs->base, ret);
}

static void w83793g_update_intr_reading(pp_sensor_scannable_t* s) {
    w83793g_sensor_t* wbs = PP_TP_SCANNABLE_2_OBJ_CAST(s, w83793g_sensor_t);
    sensor_config_t* cfg = wbs->cfg;
    int ret = PP_ERR;
    pp_tp_i2c_chip_t* chip = wbs->chip;
    u_char value;

    if (pp_tp_i2c_chip_is_initialized(chip)
	&& pp_tp_scan_sensdev_do_scan(&wbs->base)
	&& (PP_SUC == pp_tp_i2c_chip_pre_com(chip))) {

        ret = pp_tp_i2c_chip_rx_byte_data(chip, cfg->config.intr.reg,
					  &value);
        pp_tp_i2c_chip_post_com(chip);

        if (ret == PP_SUC) {
	    // check if intr bit is set
	    ret = value & (1 << cfg->config.intr.bit) ? 1 : 0;
        }
	else
	   ret = -2;
    }
    pp_tp_scan_sensdev_update_reading(&wbs->base, ret);
}

static void w83793g_destroy(w83793g_sensor_t* sensor) {
    pp_tp_scan_sensdev_unregister(&sensor->base);
    pp_tp_i2c_chip_release(sensor->chip);
    pp_tp_scan_sensdev_cleanup(&sensor->base);
    free(sensor);
}


