#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <getopt.h>
#include <asm/io.h>
#include <pp/i2c.h>
#include <pp/bmc/bmc.h>

#include <pp/base.h>
#include <pp/cfg.h>
#include <pp/bmc/bmc_dev_oem_asus.h>

#define DEFAULT_I2C_FILENAME	"/dev/i2c-faraday-4"

static const char *usage_msg =
"\n"
"   Options:\n"
"     -f         ..  flush the i2c slave buffer\n"
"     -u <file>  ..  Update ASUS SEEPROM with new topology file\n"
"\n";

int main(int argc, char **argv)
{
	unsigned char	write_buf[TOP_FILE_MAX_LENGTH];
	unsigned char	write_buf_act[2 + 64];		// 2 offset bytes
	int		i2c_slave_addr = 0xa0;
	int		fpstart = 0, fpend = 0;
	unsigned short	write_length = 0, write_length_left, write_offset, write_cnt, seeprom_offset;
	unsigned short	file_length, file_checksum = 0, checksum, idx;
    int		fd, c, n_fd;
    FILE	*fp;
    size_t	n;
    char	filename[] = DEFAULT_I2C_FILENAME;
    char	*ufilename = NULL;
	int		do_flush = 0;

    while ((c = getopt(argc, argv, "fu:")) != -1) {
		switch (c) {
			case 'u':
		  		ufilename = strdup(optarg);
				break;
			case 'f':
				do_flush = 1;
				break;
			case '?':
//				printf("Unknown option found !\n");;
		      	break;
		};
	};
    if (ufilename == NULL) {
		printf(usage_msg);
		return (0x10);
    };
 	
	printf("\n");

    /* Read Topology file */
    if ((fp = fopen(ufilename, "rb")) == NULL){
		/* ERROR HANDLING; you can check errno to see what went wrong */
       	fprintf(stderr, "[ERROR] Error Opening update ASUS topology file %i.\n",errno);
       	exit(0x20);
    };
    if (fseek(fp, 0, SEEK_SET) == 0) {
    	fpstart = ftell(fp);
    };
    if (fseek(fp, 0, SEEK_END) == 0) {
    	fpend = ftell(fp);
    };
    if ((fpstart != -1) && (fpend != -1) && (fpend > fpstart)) {
    	write_length = fpend - fpstart;
   	    fseek(fp, 0, SEEK_SET);
   	    if (write_length <= (TOP_FILE_MAX_LENGTH - 4)) {
   			//byte 0: file length LSB	(include "file length", "file checksum")
   			//byte 1: file length MSB
   			//byte 2: file checksum LSB	(include "file length", but exclude "file checksum")
   			//byte 3: file checksum MSB
// File length check
		    n = fread(&file_length, 2, 1, fp);
		    if ((n != 1)) {
		       	fprintf(stderr, "[ERROR] Topology file read 'file length' fail\n");
		       	fclose(fp);
		       	exit(0x21);
		    };
		    if (file_length != write_length) {
		       	fprintf(stderr, "[ERROR] Topology file 'file length' not match\n");
		       	fclose(fp);
		       	exit(0x22);
		    };
			write_buf[0] = (unsigned char)(write_length & 0x00FF);			// File length LSB
			write_buf[1] = (unsigned char)((write_length & 0xFF00) >> 8);	// File length MSB
		    n = fread(&(write_buf[2]), (write_length - 2), 1, fp);
		    if ((n != 1)) {
		       	fprintf(stderr, "[ERROR] Topology file read fail\n");
		       	fclose(fp);
		       	exit(0x23);
		    };
// File checksum check
		    file_checksum = *((unsigned short *)(&write_buf[2]));
		    checksum = write_buf[0] + write_buf[1];
		    for (idx = 0; idx < (write_length - 4); idx++) {
		    	checksum += write_buf[idx + 4];
		    };
		    checksum += file_checksum;
		    if (checksum != 0) {
		       	fprintf(stderr, "[ERROR] Topology file file checksum fail\n");
		       	fclose(fp);
//		       	fprintf(stderr, "Checksum = 0x%4.4X\n", checksum);
		       	exit(0x24);
		    };
	       	printf("Topology file read in buffer success (%d bytes)\n", (write_length - 4));
			fclose(fp);
   	    } else {
	       	fprintf(stderr, "[ERROR] Topology file length exceeded maximum %d bytes\n", (TOP_FILE_MAX_LENGTH - 4));
	       	fclose(fp);
	       	exit(0x25);
   	    };
    } else {
       	fprintf(stderr, "[ERROR] Topology file length uncorrect\n");
       	fclose(fp);
       	exit(0x26);
	};

    /* Open the I2C interface */
    if ((fd = open(filename,O_RDWR)) < 0) {
		fprintf(stderr, "[ERROR] Error Opening I2C device %i.\n",errno);
		exit(0x30);
    };
    /* flush slave buffer */
    if (do_flush) {
    	if (ioctl(fd, I2C_SLAVE_FLUSH_BUFFER, 0) < 0) {
    	    fprintf(stderr, "[ERROR] Error flushing i2c buffer: %i.\n", errno);
	       	close(fd);
	       	exit(0x31);
    	} else {
    	    printf("i2c buffer flushed.\n");
    	};
    	usleep(500000);
    };
    /* set I2C slave address */
   	if (ioctl(fd, I2C_SLAVE_FORCE, (i2c_slave_addr >> 1)) < 0) {
   	    fprintf(stderr, "[ERROR] Error setting I2C slave address: %i.\n", errno);
       	close(fd);
       	exit(0x32);
   	} else {
   	    printf("I2C slave address set to 0x%02x.\n", i2c_slave_addr);
   	};
    /* Write topology file to SEEPROM device */		//Page Write : 64 bytes
    write_length_left = write_length + 2;
    write_offset = 0;
    seeprom_offset = TOP_FILE_SEEPROM_OFFSET;
    while (write_length_left > 0) {
 		write_cnt = (write_length_left >= 64) ? 64 : write_length_left;
    	memcpy(&(write_buf_act[0]), (unsigned char *)(&seeprom_offset) + 1, 1);	//MSB offset
    	memcpy(&(write_buf_act[1]), (unsigned char *)(&seeprom_offset), 1);		//LSB offset
    	memcpy(&(write_buf_act[2]), &(write_buf[write_offset]), write_cnt);
	    n_fd = write(fd, write_buf_act, (write_cnt + 2));
	    if ((n_fd < 0) || (n_fd != (write_cnt + 2))) {
	   	    fprintf(stderr, "\n[ERROR] Error writing topology file to SEEPROM at offset %d : %i.\n", write_offset, errno);
	       	close(fd);
	       	exit(0x33);
	    };
	    write_length_left -= write_cnt;
	    write_offset += write_cnt;
	    seeprom_offset += write_cnt;
		printf("\rWrite done %3d%%", (write_offset * 100 / (write_length + 2)));
		fflush(stdout);
	    usleep(20000);
	};
    printf("\nWrite the ASUS topology file to SEEPROM success !\n");
	close(fd);

	return (0x00);
}


