/*
 * <3BSD notice header>
 *
 * Copyright (C) 2003-2008 Raritan, Inc. <http://www.raritan.com/>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 * 
 *     * Neither the name of Raritan, Inc. nor the names of
 *       its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written
 *       permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <pp/firmware.h>
#include <pp/intl.h>
#include <liberic_misc.h>
#include <pp/um.h>
#if defined(PRODUCT_FLASHX4)
# include <pp/flashprog.h>
#endif
#include "eric_base.h"
#include "eric_util.h"
#include "eric_validate.h"
#include "eric_forms.h"
#include "wsIntrn.h"
#include "tmpl_firmware_common.h"
#if defined(PP_FEAT_RPCCFG)
# include <pp/rpc_ripc.h>
#endif

#include "tmpl_firmware_upload.h"


FV_SPEC = {
};

static int pre_validate_hook(webs_t wp, form_handler_t * fh);
static int fw_upload_cb(webs_t wp, char * name, char * filename, void * data, size_t data_offset, size_t data_len, int more_data);

#ifdef ASUS_FLASH_EEPROM
static unsigned int asus_fw_length = 0;		//First 4 bytes in image file
unsigned int asus_cfg_length = 0;			//Total image size - F/W size - 4, global variable
#endif


int
firmware_upload_tmpl_init(void)
{
    form_handler_t * fh;

    if (firmware_common_tmpl_init() != 0) return -1;

    /* register POST callbacks */
    websAddPostDataCallback("firmware_file", fw_upload_cb);

    fh = CREATE_FH_INSTANCE(TEMPLATE_FIRMWARE_UPLOAD, ACL_OBJ_FIRMWARE);

    fh->pre_validate_hook = pre_validate_hook;

    REGISTER_FH_INSTANCE_AND_RETURN(fh);
}

static int
pre_validate_hook(webs_t wp,  form_handler_t * fh UNUSED)
{
#if defined(PRODUCT_FLASHX4)
    int  file_type = 0;
    char *fp_upload;
#endif    
    char logmsg[PP_NOTIFY_MAX_MSG_LEN+1];
    const char * fCrossOEM = websGetVar(wp, "crossoem", "");
    const char * fCrossHWID = websGetVar(wp, "crosshwid", "");
#if defined(PP_FEAT_RPCCFG)
    const char * fIgnoreSlave = websGetVar(wp, "ignoreslave", "");
#endif
    const char * error_msg = NULL;
    char * ver = NULL;
    char * min_req_ver = NULL;
    char * min_dwngrd_ver = NULL;
    int bnr;
    int ret = -1;

    if (wp->upload_failed) {
	return -1;
    } else if (form_button_clicked(wp, "action_firmware_upload")) {
	wp->fh_flags |= FH_FLAG_ABORT_AT_VALIDATE;

#if defined(PRODUCT_FLASHX4)
	fp_upload = websGetVar(wp, "upload_file", "");

	if ( !strcmp(fp_upload, "8mb") ) {
	    file_type = FILE_UBOOT_8MB;
	} else if ( !strcmp(fp_upload, "16mb") ) {
	    file_type = FILE_UBOOT_16MB;
	} else {
#endif /* PRODUCT_FLASHX4 */

	    /* get firmware info */
	    if (PP_FAILED(pp_firmware_erla_get_info(wp->sd->firmware_ctx, &ver, &bnr,
			  NULL, &min_req_ver, &min_dwngrd_ver))) {
		set_response(wp, ERIC_RESPONSE_ERROR, _("Getting firmware information failed."));
		goto bail;
	    }

	    if (pp_firmware_erla_validate(wp->sd->firmware_ctx, !strcmp(fCrossOEM, "yes"),
					  !strcmp(fCrossHWID, "yes")) != 0) {
		char * custom_err_msg = NULL;
		if (errno == PP_FIRMWARE_ERR_NEW_VER_TOO_OLD) {
		    asprintf(&custom_err_msg,
			     _("You may not downgrade to the uploaded firmware. "
			       "The minimum version you may downgrade to is %c%c.%c%c.%c%c."),
			     min_dwngrd_ver[0], min_dwngrd_ver[1], min_dwngrd_ver[2],
			     min_dwngrd_ver[3], min_dwngrd_ver[4], min_dwngrd_ver[5]);
		} else if (errno == PP_FIRMWARE_ERR_CUR_VER_TOO_OLD) {
		    asprintf(&custom_err_msg,
			     _("The current firmware is too old. "
			       "You need to update to version %c%c.%c%c.%c%c first."),
			     min_req_ver[0], min_req_ver[1], min_req_ver[2],
			     min_req_ver[3], min_req_ver[4], min_req_ver[5]);
		} else {
		    error_msg = pp_error_string(errno);
		}
		set_response(wp, ERIC_RESPONSE_ERROR, custom_err_msg ? custom_err_msg : error_msg);
		free(custom_err_msg);
		eric_notify_post_event(_("Firmware validation failed."), "device", PP_NOTIFY_EVENT_GENERIC);
	    } else {
		snprintf(logmsg, sizeof(logmsg),
			 _("Firmware file uploaded by user '%s'. %c%c.%c%c.%c%c (Build %u)."),
			 eric_session_get_user(wp->session),
			 ver[0], ver[1], ver[2], ver[3], ver[4], ver[5], bnr);
		eric_notify_post_event(logmsg, "device", PP_NOTIFY_EVENT_GENERIC);

#ifdef PP_FEAT_RPCCFG
		if (strcmp(fIgnoreSlave, "yes")) {
		    /*
		     * Send fw via RPC to the slave KIM. This is only executed by the master KIM
		     * because no web-access to slave KIMs.
		     */
		    if (PP_FAILED(pp_rpc_fw_upload(wp->sd->firmware_ctx))) {
			set_response(wp, ERIC_RESPONSE_ERROR,
				     _("A internal error occured during sharing the firmware internally."));
			goto bail;
		    }
		}
#endif /* PP_FEAT_RPCCFG */
		ret = 0;
	    }
#if defined(PRODUCT_FLASHX4)
	}
	if (file_type) {
	    int ret;
	    size_t size = pp_firmware_erla_data_get_len(wp->sd->firmware_ctx);
	    if (size == 0x20000) {
		void * flash_data = malloc(size);
		if ((ret = pp_firmware_erla_data_get(wp->sd->firmware_ctx, flash_data, 0, size)) != 0
		    || (ret = pp_flashprog_set_data(flash_data, size, file_type)) != 0) {
		    set_response(wp, ERIC_RESPONSE_ERROR, _("Error while uploading flash prog file."));
		} else {
		    set_response(wp, ERIC_RESPONSE_OK, _("Flash file successfully uploaded"));
		}
		free(flash_data);
	    } else {
		set_response(wp, ERIC_RESPONSE_ERROR, _("Flash file has wrong size: 0x%x. It should have 0x20000 bytes."), size);
	    }
	    pp_firmware_erla_ctx_free(wp->sd->firmware_ctx);
	    wp->sd->firmware_ctx = 0;
	    return ret;
	}
#endif /* PRODUCT_FLASHX4 */

    } else {
	ret = 0;
    }

 bail:
    free(ver);
    free(min_req_ver);
    free(min_dwngrd_ver);
    if (ret != 0) {
	pp_firmware_erla_ctx_free(wp->sd->firmware_ctx);
	wp->sd->firmware_ctx = 0;
    }
    return ret;
}

#ifdef ASUS_FLASH_EEPROM

static int
fw_upload_cb(webs_t wp, char * name UNUSED, char * filename UNUSED,
	     void * data, size_t data_offset, size_t data_len, int more_data)
{
	char *data_new = (char *)data;
	size_t data_len_new = data_len;
	size_t data_offset_new = data_offset;
	size_t asus_cfg_length_new = 0;
	char *asus_cfg_data_new = (char *)data;
	static unsigned char asus_cfg_file_opened = 0;
    int fd, n;
	
    eric_session_touch(wp->session);

    if (PP_ERR == pp_um_user_has_permission(wp->user, ACL_OBJ_FIRMWARE, pp_acl_raasip_yes_str)) {
		set_response(wp, ERIC_RESPONSE_ERROR, perm_denied_msg);
		goto fail;
    };

    if (data_offset == 0 && data_len > 0) {
	if ((wp->sd->firmware_ctx = pp_firmware_erla_ctx_new(PP_FIRMWARE_ERLA_MAX_SIZE+2, 1)) == 0) {
	    set_response(wp, ERIC_RESPONSE_ERROR, _("A firmware upload/update is already running by another user."));
	    goto fail;
		};
		asus_fw_length = *((unsigned int *)data);
		pp_log("%s(): ASUS F/W length = %d\n", ___F, asus_fw_length);
		asus_cfg_file_opened = 0;
		asus_cfg_length = 0;
    };

//	pp_log("%s(): Upload F/W image in offset = %d\n", ___F, data_offset);
//	pp_log("%s(): Upload F/W image in length = %d\n", ___F, data_len);
	if (data_offset_new == 0) {
		if (data_len_new >= sizeof(unsigned int)) {
			data_new += sizeof(unsigned int);
			data_len_new -= sizeof(unsigned int);
		} else {
			data_len_new = 0;
		};
	} else if (data_offset_new < sizeof(unsigned int)) {
		if ((data_offset_new + data_len_new) > sizeof(unsigned int)) {
			data_offset_new = 0;
			data_new += sizeof(unsigned int) - data_offset_new;
			data_len_new = data_len_new - (sizeof(unsigned int) - data_offset_new);
		} else {
			data_len_new = 0;
		};
	} else {
		data_offset_new -= sizeof(unsigned int);
	};
	if (data_offset_new < asus_fw_length) {
		if ((data_offset_new + data_len_new) > asus_fw_length) {
			data_len_new = asus_fw_length - data_offset_new;
			if (!more_data) {
				asus_cfg_length_new = data_len - data_len_new - 2;
			} else {
				asus_cfg_length_new = data_len - data_len_new;
			};
			asus_cfg_length += asus_cfg_length_new;
			asus_cfg_data_new += data_len_new;
		};
	    if (pp_firmware_erla_data_set(wp->sd->firmware_ctx, (void *)data_new, data_offset_new, data_len_new) != 0) {
			set_response(wp, ERIC_RESPONSE_ERROR, _("Firmware too big (max. %u MB)."), PP_FIRMWARE_ERLA_MAX_SIZE / (1024 * 1024));
			goto fail;
		};
//		pp_log("%s(): Write done F/W image out offset = %d\n", ___F, data_offset_new);
//		pp_log("%s(): Write done F/W image out length = %d\n", ___F, data_len_new);
	} else {
		if (!more_data) {
			asus_cfg_length_new = data_len - 2;
		} else {
			asus_cfg_length_new = data_len;
		};
		asus_cfg_length += asus_cfg_length_new;
	};

	if (asus_cfg_length_new) {
	    assert(ASUS_TOP_CFG_FILE_PATH);
		if (asus_cfg_file_opened) {
		    /* open file -- Append */
		    if (0 > (fd = open(ASUS_TOP_CFG_FILE_PATH, O_WRONLY | O_APPEND, S_RW_RW_RW_))) {
		        pp_bmc_log_error("%s can't append to file %s:", ___F, ASUS_TOP_CFG_FILE_PATH);
		        goto fail;
		    };
		} else {
		    /* open file -- Creat new (first time only) */
		    if (0 > (fd = open(ASUS_TOP_CFG_FILE_PATH, O_WRONLY | O_TRUNC | O_CREAT, S_RW_RW_RW_))) {
		        pp_bmc_log_error("%s can't open new file %s:", ___F, ASUS_TOP_CFG_FILE_PATH);
		        goto fail;
		    };
		    asus_cfg_file_opened = 1;
		};
	    /* write file */
	    n = write(fd, asus_cfg_data_new, asus_cfg_length_new);
	    if ((n < 0) || (n != (int)asus_cfg_length_new)) {
	        pp_bmc_log_error("%s write to file %s error", ___F, ASUS_TOP_CFG_FILE_PATH);
	        close(fd);
	        goto fail;
	    };
        close(fd);
	};

    if (data == NULL) {
		pp_firmware_erla_ctx_free(wp->sd->firmware_ctx);
		wp->sd->firmware_ctx = 0;
    } else if (!more_data) {
//		if ((data_offset + data_len) >= 2)
//			pp_firmware_erla_data_set_len(wp->sd->firmware_ctx, data_offset + data_len - 2);
		pp_firmware_erla_data_set_len(wp->sd->firmware_ctx, asus_fw_length);
		pp_log("%s(): ASUS top cfg length = %d\n", ___F, asus_cfg_length);
		if (asus_cfg_length == 0 || asus_cfg_length != (data_offset + data_len - 2 - asus_fw_length - sizeof(unsigned int))) {
			pp_log("%s(): ASUS top cfg length not matched !\n", ___F);
			goto fail;
		};
    };
  
    return 0;

 fail:
    pp_firmware_erla_ctx_free(wp->sd->firmware_ctx);
    wp->sd->firmware_ctx = 0;
    wp->upload_failed = 1;
    eric_notify_post_event(_("Firmware upload failed."), "device", PP_NOTIFY_EVENT_GENERIC);

    return -1;
}

#else

static int
fw_upload_cb(webs_t wp, char * name UNUSED, char * filename UNUSED,
	     void * data, size_t data_offset, size_t data_len, int more_data)
{
    eric_session_touch(wp->session);

    if (PP_ERR == pp_um_user_has_permission(wp->user, ACL_OBJ_FIRMWARE,
                                            pp_acl_raasip_yes_str)) {
	set_response(wp, ERIC_RESPONSE_ERROR, perm_denied_msg);
	goto fail;
    }

    if (data_offset == 0 && data_len > 0) {
	if ((wp->sd->firmware_ctx = pp_firmware_erla_ctx_new(PP_FIRMWARE_ERLA_MAX_SIZE+2, 1)) == 0) {
	    set_response(wp, ERIC_RESPONSE_ERROR, _("A firmware upload/update is already running by another user."));
	    goto fail;
	}
    }

    if (pp_firmware_erla_data_set(wp->sd->firmware_ctx, data, data_offset, data_len) != 0) {
	set_response(wp, ERIC_RESPONSE_ERROR, _("Firmware too big (max. %u MB)."), PP_FIRMWARE_ERLA_MAX_SIZE / (1024 * 1024));
	goto fail;
    }

    if (data == NULL) {
	pp_firmware_erla_ctx_free(wp->sd->firmware_ctx);
	wp->sd->firmware_ctx = 0;
    } else if (!more_data) {
	if ((data_offset + data_len) >= 2) pp_firmware_erla_data_set_len(wp->sd->firmware_ctx, data_offset + data_len - 2);
    }
  
    return 0;

 fail:
    pp_firmware_erla_ctx_free(wp->sd->firmware_ctx);
    wp->sd->firmware_ctx = 0;
    wp->upload_failed = 1;
    eric_notify_post_event(_("Firmware upload failed."), "device", PP_NOTIFY_EVENT_GENERIC);

    return -1;
}


#endif
