/*
 * <3BSD notice header>
 *
 * Copyright (C) 2003-2008 Raritan, Inc. <http://www.raritan.com/>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 * 
 *     * Neither the name of Raritan, Inc. nor the names of
 *       its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written
 *       permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <liberic_config.h>
#include <pp/intl.h>
#include <pp/ipmi.h>
#include "ej.h"
#include "eric_base.h"
#include "eric_util.h"
#include "eric_forms.h"
#include "eric_form_vars.h"
#include "eric_validate.h"
#include <pp/vsc.h>


FV_SPEC = {
    /* we have no form vars (yet) */
};

static int pre_validate_hook(webs_t wp, form_handler_t * fh);
static int ipmi_sensors_build_table_asp(int eid, webs_t wp UNUSED, int argc UNUSED, char **argv UNUSED);

int ipmi_sensors_tmpl_init(void)
{
    form_handler_t * fh;

    /* register ASPs */
    websAspDefine("ipmiGetSensorTable", ipmi_sensors_build_table_asp);

    fh = CREATE_FH_INSTANCE(TEMPLATE_IPMI_SENSORS, ACL_OBJ_IPMI_STATUS);
    fh->pre_validate_hook = pre_validate_hook;

    REGISTER_FH_INSTANCE_AND_RETURN(fh);
}

/* the formular handler (parses the buttons) */
static int pre_validate_hook(webs_t wp, form_handler_t * fh UNUSED)
{
    if (form_button_clicked(wp, "action_refresh")) {
            /* dont show "Operation Completed" */
            set_response(wp, ERIC_RESPONSE_OK, "");
    }

    return 0;
}

/* ASP calls */
static const char* fixup_string(char *s) {
    pp_trim_string(s);
    if (!s || !s[0]) {
        return "&nbsp;";
    }
    return s;
}

#define TABLE_BG_WHITE  0
#define TABLE_BG_RED    1
#define TABLE_BG_GREEN  2
#define TABLE_BG_YELLOW 3

static void add_sensor_table_entry(pp_strstream_t *sensor_table, int idx, int color,
                                   char *type, char *name, char *status, char *reading) {
#if !defined(PP_FEAT_RARITAN_DESIGN)
    const char *colors[] = { "#FFFFFF", "#FF0000", "#00FF00", "#FFFF00" };
    pp_strappend(sensor_table,
            idx % 2 ? "<tr class=\"loggingRowOdd\">\n" : "<tr class=\"loggingRowEven\">\n");

    if (color != TABLE_BG_WHITE) {
        pp_strappend(sensor_table, "  <td style=\"background-color:");
        pp_strappend(sensor_table, colors[color]);
        pp_strappend(sensor_table, "\">&nbsp;&nbsp;&nbsp;</td>\n");
    } else {
        pp_strappend(sensor_table, "<td>&nbsp;&nbsp;&nbsp;</td>\n");
    }

    pp_strappend(sensor_table, "  <td>");
    pp_strappend(sensor_table, fixup_string(type));
    pp_strappend(sensor_table, "</td>\n");

    pp_strappend(sensor_table, "  <td>");
    pp_strappend(sensor_table, fixup_string(name));
    pp_strappend(sensor_table, "</td>\n");

    pp_strappend(sensor_table, "  <td>");
    pp_strappend(sensor_table, fixup_string(status));
    pp_strappend(sensor_table, "</td>\n");

    pp_strappend(sensor_table, "  <td>");
    pp_strappend(sensor_table, fixup_string(reading));
    pp_strappend(sensor_table, "</td>\n");

    pp_strappend(sensor_table, "</tr>\n");
#else /* PP_FEAT_RARITAN_DESIGN */
    (void)idx;
    (void)color;
    pp_strappend(sensor_table, "<script>table_entry('");
    pp_strappend(sensor_table, fixup_string(type));

    pp_strappend(sensor_table, "','");
    pp_strappend(sensor_table, fixup_string(name));

    pp_strappend(sensor_table, "','");
    pp_strappend(sensor_table, fixup_string(status));

    pp_strappend(sensor_table, "','");
    pp_strappend(sensor_table, fixup_string(reading));
    pp_strappend(sensor_table, "');</script>\n");
#endif /* PP_FEAT_RARITAN_DESIGN */
}

static void add_empty_table_entry(pp_strstream_t *sensor_table, const char *str) {
    pp_strappend(sensor_table, "\n");
#if !defined(PP_FEAT_RARITAN_DESIGN)
    pp_strappend(sensor_table, "<tr><td align=\"center\" class=\"normal\">");
    pp_strappend(sensor_table, str);
    pp_strappend(sensor_table, "</td></tr>\n");
#else /* PP_FEAT_RARITAN_DESIGN */
    pp_strappend(sensor_table, "<script>table_entry('");
    pp_strappend(sensor_table, str);
    pp_strappend(sensor_table, "','','','');</script>\n");
#endif /* PP_FEAT_RARITAN_DESIGN */
}

static int get_color_from_status(pp_ipmi_sdr_list_entry_t *entry) {
    switch (entry->data.full.status) {
        case PP_IPMI_SENSOR_STATUS_NO_READING:
            return TABLE_BG_WHITE;
        case PP_IPMI_SENSOR_STATUS_OK:
            return TABLE_BG_GREEN;
        case PP_IPMI_SENSOR_STATUS_UPPER_NON_CRITICAL:
        case PP_IPMI_SENSOR_STATUS_LOWER_NON_CRITICAL:
            return TABLE_BG_YELLOW;
        default:
            return TABLE_BG_RED;
    }
}

static int ipmi_sensors_build_table_asp(int eid, webs_t wp, int argc UNUSED, char **argv UNUSED) {
    pp_ipmi_return_t ipmi_ret;
    pp_strstream_t sensor_table = PP_STRSTREAM_INITIALIZER;
    u_int i, sensor_cnt = 0, retry;

    memset(&ipmi_ret, 0, sizeof(ipmi_ret));
    pp_strstream_init(&sensor_table);

    /* if error, try once more */
    for(retry=0;retry<2;retry++){
        if (0 == pp_ipmi_send_command(PP_IPMI_CMD_SDR, PP_IPMI_SDR_SUBCMD_LIST,
			     NULL, &ipmi_ret, NULL, wp->user)){
            break;
        }
    }
    if(retry >= 2){
        add_empty_table_entry(&sensor_table, _("Error: Could not query Sensor repository."));
	goto bail;
    }

    sensor_cnt = vector_size(ipmi_ret.data.sdr_list);
    //pp_log("found %d SDR entries.\n", sensor_cnt);

    if (sensor_cnt == 0) {
	add_empty_table_entry(&sensor_table, _("No sensors found."));
	goto bail;
    }

    for (i = 0; i < sensor_cnt; i++) {
	pp_ipmi_sdr_list_entry_t *entry = vector_get(ipmi_ret.data.sdr_list, i);
	int tolerance_valid = entry->data.full.tolerance.valid;
	float tolerance = entry->data.full.tolerance.value;
	float reading = entry->data.full.reading.value;
	int tolerance_is_0, tolerance_prec, reading_prec;
	int color = TABLE_BG_WHITE;
	char reading_str[50] = "";
	char *disp_reading_str = reading_str;
	char *status = NULL;

	if (!entry) {
	    pp_log("%s(): Error: could not read entry %u\n", ___F, i);
	    continue;
	}

	if (entry->type == PP_IPMI_SENSOR_TYPE_FULL) {
	    if (entry->data.full.reading_present == PP_IPMI_SENSOR_READING_OK) {
		tolerance_is_0 = (int)(tolerance * 1000 + 0.5) == 0;
		tolerance_prec = ((int)(tolerance * 1000 + 0.5) % 1000) ? 3 : 0;
		reading_prec = ((int)(reading * 1000 + 0.5) % 1000) ? 3 : 0; 
		color = get_color_from_status(entry);
		status = strdup(pp_strstream_buf(&entry->data.full.status_string));
		if (entry->data.full.type == PP_IPMI_SENSOR_FULL_TYPE_ANALOG) {
		    if (!tolerance_valid || tolerance_is_0) {
			snprintf(reading_str, sizeof(reading_str), "%.*f %s\n",
			         reading_prec, reading,
				 entry->data.full.unit.buf ? : "");
		}    else {
			snprintf(reading_str, sizeof(reading_str), "%.*f (+/- %.*f) %s\n",
			         reading_prec, reading, tolerance_prec, tolerance,
				 entry->data.full.unit.buf ? : "");
		    }
		} // else: discrete sensors don't have a reading, just a status
	    } else {		
		status = strdup("No reading");
	    }
	} else if (entry->type == PP_IPMI_SENSOR_TYPE_COMPACT_SENSOR) {
	    if (entry->data.compact.status == PP_IPMI_SENSOR_STATUS_NO_READING) {
		status = strdup("No reading");
	    } else if (!entry->data.compact.scanning_disabled) {
		status = pp_strdup_safe(entry->data.compact.description.buf);
	    }
	} else {
	    continue;
	}
	add_sensor_table_entry(&sensor_table, i, color, entry->sdr_type_sub.buf,
			       entry->sdr_name.buf, status, disp_reading_str);
	free(status);
    }

 bail:
    ejSetResult(eid, sensor_table.buf);

    pp_strstream_free(&sensor_table);
    pp_ipmi_cleanup_ret(&ipmi_ret);

    return 0;
}
